#include "Solver.h"

#include <assert.h>
#include <iostream>

#include "../Pattern/Pattern.h"

namespace
{
	const double SmallValue = 1e-15;

	void ZeroIze(double& value)
	{
		if (value < SmallValue && value > -SmallValue)
		{
			value = 0;
		}
	}
}

NoThresholdSolver::NoThresholdSolver(size_t patternSize, bool alwaysSolve)
	: m_PatternSize(patternSize)
	, m_SolveAlways(alwaysSolve)
{
	InitMatrix();
}

void NoThresholdSolver::InitMatrix()
{
	std::vector<double> initVector(m_PatternSize + 1, 0);
	m_Matrix.assign(m_PatternSize, initVector);
	m_Solution.assign(m_PatternSize, 0);
}

void NoThresholdSolver::FeedPatterns(const std::vector<Pattern>& patterns)
{
	size_t count = patterns.size();

	for (size_t i = 0; i < count; ++i)
	{
		const Pattern& pattern = patterns[i];

		if (pattern.Size() == m_PatternSize)
		{
			FeedPattern(pattern);
		}
	}
}

void NoThresholdSolver::FeedPattern(const Pattern& pattern)
{
	size_t size = pattern.Size();

	assert(size == m_PatternSize);
	if (size != m_PatternSize)
	{
		return;
	}

	const std::vector<double>& values = pattern.Values();
	double etalon = pattern.Etalon();

	size_t i;

	for (i = 0; i < size; ++i)
	{
		for (size_t j = 0; j < size; ++j)
		{
			m_Matrix[i][j] += values[i]*values[j];
		}
		m_Matrix[i][size] += values[i]*etalon;
	}
}

const std::vector<double>& NoThresholdSolver::Solve()
{
	if (m_Matrix.empty())
	{
		InitMatrix();
		return m_Solution;
	}
	
	size_t iSize = m_Matrix.size();
	size_t jSize = m_Matrix[0].size();

	assert(iSize + 1 == jSize);
	assert(jSize == m_PatternSize + 1);

	Print();
    
	size_t i, j;

	for (i = 0; i < iSize; ++i)
	{
		j = i;

		while (m_Matrix[j][i] == 0)
		{
			if (++j == iSize)
			{
				if (m_SolveAlways)
				{
					size_t idx = FindZeroLine();
					if (idx < iSize)
					{
						j = idx;
						m_Matrix[j][i] = 1;
					}
					else return m_Solution;
				}
				else return m_Solution;
			}
		}

		Divide(i, m_Matrix[i][i]);

		if ( j != i)
		{
			std::swap(m_Matrix[i], m_Matrix[j]);
		}

		for (j = i + 1; j < iSize; ++j)
		{
			Substract(i, j, m_Matrix[j][i]);
		}
		Print();
	}

	for (i = iSize - 1; i > 0; --i)
	{
		for (j = i; j > 0; --j)
		{
			Substract(i, j - 1, m_Matrix[j - 1][i]);
		}
	}

	Print();

	std::vector<double> result(iSize);

	for (i = 0; i < iSize; ++i)
	{
		result[i] = m_Matrix[i][iSize];
	}
	m_Solution = result;

	return m_Solution;
}

const std::vector<double>& NoThresholdSolver::Solution() const
{
	return m_Solution;	
}

void NoThresholdSolver::Substract(size_t whatIdx, size_t fromIdx, double coef)
{
	size_t jSize = m_Matrix[whatIdx].size();

	for (size_t j = 0; j < jSize; ++j)
	{
		m_Matrix[fromIdx][j] -= m_Matrix[whatIdx][j]*coef;
		ZeroIze(m_Matrix[fromIdx][j]);
	}
}

void NoThresholdSolver::Divide(size_t whatIdx, double coef)
{
	size_t jSize = m_Matrix[whatIdx].size();

	for (size_t j = 0; j < jSize; ++j)
	{
		m_Matrix[whatIdx][j] /= coef;
	}
}

void NoThresholdSolver::Print()
{
/*	for (size_t i = 0; i < m_Matrix.size(); ++i)
	{
		for (size_t j = 0; j < m_Matrix[i].size(); ++j)
		{
			std::cout << m_Matrix[i][j] << "\t";
		}
		std::cout << "\n";
	}
	std::cout << "\n";*/
}

size_t NoThresholdSolver::FindZeroLine() const
{
	size_t iSize = m_Matrix.size();
	size_t jSize = m_Matrix[0].size() - 1;

	for (size_t i = 0; i < iSize; ++i)
	{
		bool isZero = true;
		for (size_t j = 0; j < jSize; ++j)
		{
			if (m_Matrix[i][j] != 0)
			{
				isZero = false;
				break;
			}
		}
		if (isZero)
		{
			return i;
		}
	}
	return iSize;
}
