#include "Solver.h"

#include <assert.h>
#include <iostream>

#include "../Pattern/Pattern.h"

namespace
{
	const double SmallValue = 1e-10;

	void ZeroIze(double& value)
	{
		if (value < SmallValue && value > -SmallValue)
		{
			value = 0;
		}
	}
}

Solver::Solver(size_t patternSize, bool alwaysSolve)
	: m_PatternSize(patternSize)
	, m_SolveAlways(alwaysSolve)
{
	InitMatrix();
}

void Solver::InitMatrix()
{
	std::vector<double> initVector(m_PatternSize + 2, 0);
	m_Matrix.assign(m_PatternSize + 1, initVector);
	m_Solution.assign(m_PatternSize + 1, 0);
}

void Solver::FeedPatterns(const std::vector<Pattern>& patterns)
{
	size_t count = patterns.size();

	for (size_t i = 0; i < count; ++i)
	{
		const Pattern& pattern = patterns[i];

		if (pattern.Size() == m_PatternSize)
		{
			FeedPattern(pattern);
		}
	}
}

void Solver::FeedPattern(const Pattern& pattern)
{
	size_t size = pattern.Size();

	assert(size == m_PatternSize);
	if (size != m_PatternSize)
	{
		return;
	}

	const std::vector<double>& values = pattern.Values();
	double etalon = pattern.Etalon();
    
	size_t i;

	for (i = 0; i < size; ++i)
	{
		for (size_t j = 0; j < size; ++j)
		{
			m_Matrix[i][j] += values[i]*values[j];
		}
		m_Matrix[i][size] += values[i];
		m_Matrix[i][size + 1] += values[i]*etalon;
	}

	for (i = 0; i < size; ++i)
	{
		m_Matrix[size][i] += values[i];
	}
	m_Matrix[size][size] += 1;
	m_Matrix[size][size + 1] += etalon;
}

const std::vector<double>& Solver::Solve()
{
	if (m_Matrix.empty())
	{
		InitMatrix();
		return m_Solution;
	}
	
	size_t iSize = m_Matrix.size();
	size_t jSize = m_Matrix[0].size();

	m_ZeroMarkers.assign(iSize, false);

	assert(iSize + 1 == jSize);
	assert(jSize == m_PatternSize + 2);

	Print();

    size_t i, j;

	std::vector<size_t> indexes(iSize);
	for (i = 0; i < iSize; ++i)
	{
		indexes[i] = (unsigned)i;
	}

	if (m_Matrix[0][0] == 0)
	{
		size_t index = 1;
		while (m_Matrix[index][index] == 0)
		{
			++index;
			if (index == iSize) break;
		}

		if (index != iSize)
		{
			std::swap(m_Matrix[0], m_Matrix[index]);
			std::swap(indexes[0], indexes[index]);

			for (i = 0; i < iSize; ++i)
			{
				std::swap(m_Matrix[i][0], m_Matrix[i][index]);
			}
		}
		else return m_Solution;
	}

	for (i = 0; i < iSize; ++i)
	{
		j = i;

		while (m_Matrix[j][i] == 0)
		{
			if (++j == iSize)
			{
				if (m_SolveAlways)
				{
					size_t idx = FindZeroLine();
					if (idx < iSize)
					{
						j = idx;
						m_Matrix[j][i] = 1;
						m_ZeroMarkers[j] = false;
					}
					else return m_Solution;
				}
				else return m_Solution;
			}
		}

		Divide(j, m_Matrix[j][i]);

		if ( j != i)
		{
			std::swap(m_Matrix[i], m_Matrix[j]);
			std::swap(indexes[i], indexes[j]);
		}

		for (j = i + 1; j < iSize; ++j)
		{
			Substract(i, j, m_Matrix[j][i]);
		}

		Print();
	}

	for (i = iSize - 1; i > 0; --i)
	{
		for (j = i; j > 0; --j)
		{
			Substract(i, j - 1, m_Matrix[j - 1][i]);
		}
	}

	Print();

	m_Solution.resize(iSize);
	for (i = 0; i < iSize; ++i)
	{
		m_Solution[indexes[i]] = m_Matrix[i][iSize];
	}

	return m_Solution;
}

const std::vector<double>& Solver::Solution() const
{
	return m_Solution;	
}

void Solver::Substract(size_t whatIdx, size_t fromIdx, double coef)
{
	size_t jSize = m_Matrix[whatIdx].size();

	bool zeroMarker = true;
	for (size_t j = 0; j < jSize - 1; ++j)
	{
		m_Matrix[fromIdx][j] -= m_Matrix[whatIdx][j]*coef;
		ZeroIze(m_Matrix[fromIdx][j]);
		zeroMarker = zeroMarker && (m_Matrix[fromIdx][j] == 0);
	}
	m_ZeroMarkers[fromIdx] = zeroMarker;

	size_t pos = jSize - 1;
	m_Matrix[fromIdx][pos] -= m_Matrix[whatIdx][pos]*coef;
	ZeroIze(m_Matrix[fromIdx][pos]);
}

void Solver::Divide(size_t whatIdx, double coef)
{
	assert(coef != 0);
	size_t jSize = m_Matrix[whatIdx].size();

	for (size_t j = 0; j < jSize; ++j)
	{
		m_Matrix[whatIdx][j] /= coef;
	}
}

void Solver::Print()
{
/*	for (size_t i = 0; i < m_Matrix.size(); ++i)
	{
		for (size_t j = 0; j < m_Matrix[i].size(); ++j)
		{
			std::cout << m_Matrix[i][j] << "\t";
		}
		std::cout << "\n";
	}
	std::cout << "\n";*/
}

size_t Solver::FindZeroLine() const
{
	size_t iSize = m_Matrix.size();
	for (size_t i = 0; i < iSize; ++i)
	{
		if (m_ZeroMarkers[i]) return i;
	}
	return iSize;
}